from types import SimpleNamespace
import depthai as dai

from ..previews import Previews


class PipelineManager:
    """
    Manager class handling different :obj:`depthai.Pipeline` operations. Most of the functions wrap up nodes creation
    and connection logic onto a set of convenience functions.
    """

    def __init__(self, openvinoVersion=None, poeQuality=100, lowCapabilities=False, lowBandwidth=False):
        self.openvinoVersion=openvinoVersion
        self.poeQuality = poeQuality
        self.lowCapabilities = lowCapabilities
        self.lowBandwidth = lowBandwidth

        #: depthai.Pipeline: Ready to use requested pipeline. Can be passed to :obj:`depthai.Device` to start execution
        self.pipeline = dai.Pipeline()
        #: types.SimpleNamespace: Contains all nodes added to the :attr:`pipeline` object, can be used to conveniently access nodes by their name
        self.nodes = SimpleNamespace()

        if openvinoVersion is not None:
            self.pipeline.setOpenVINOVersion(openvinoVersion)

    #: depthai.OpenVINO.Version: OpenVINO version which will be used in pipeline
    openvinoVersion = None
    #: int, Optional: PoE encoding quality, can decrease frame quality but decrease latency
    poeQuality = None
    #: bool: If set to :code:`True`, manager will MJPEG-encode the packets sent from device to host to lower the bandwidth usage. **Can break** if more than 3 encoded outputs requested
    lowBandwidth = False
    #: bool: If set to :code:`True`, manager will try to optimize the pipeline to reduce the amount of host-side calculations (useful for RPi or other embedded systems)
    lowCapabilities = False

    _depthConfig = dai.StereoDepthConfig()
    _rgbConfig = dai.CameraControl()
    _leftConfig = dai.CameraControl()
    _rightConfig = dai.CameraControl()

    _depthConfigInputQueue = None
    _rgbConfigInputQueue = None
    _leftConfigInputQueue = None
    _rightConfigInputQueue = None

    def setNnManager(self, nnManager):
        """
        Assigns NN manager. It also syncs the pipeline versions between those two objects

        Args:
            nnManager (depthai_sdk.managers.NNetManager): NN manager instance
        """
        self.nnManager = nnManager
        if self.openvinoVersion is None and self.nnManager.openvinoVersion:
            self.pipeline.setOpenVINOVersion(self.nnManager.openvinoVersion)
        else:
            self.nnManager.openvinoVersion = self.pipeline.getOpenVINOVersion()

    def createDefaultQueues(self, device):
        """
        Creates default queues for config updates

        Args:
            device (depthai.Device): Running device instance
        """

        if hasattr(self.nodes, "stereo"):
            self._depthConfigInputQueue = device.getInputQueue("stereoConfig")
        if hasattr(self.nodes, "camRgb"):
            self._rgbConfigInputQueue = device.getInputQueue(Previews.color.name + "_control")
        if hasattr(self.nodes, "monoLeft"):
            self._leftConfigInputQueue = device.getInputQueue(Previews.left.name + "_control")
        if hasattr(self.nodes, "monoRight"):
            self._rightConfigInputQueue = device.getInputQueue(Previews.right.name + "_control")

    def closeDefaultQueues(self):
        """
        Creates default queues for config updates

        Args:
            device (depthai.Device): Running device instance
        """

        if self._depthConfigInputQueue is not None:
            self._depthConfigInputQueue.close()
        if self._rgbConfigInputQueue is not None:
            self._rgbConfigInputQueue.close()
        if self._leftConfigInputQueue is not None:
            self._leftConfigInputQueue.close()
        if self._rightConfigInputQueue is not None:
            self._rightConfigInputQueue.close()

    def __calcEncodeableSize(self, sourceSize):
        w, h = sourceSize
        if w % 16 > 0:
            newW = w - (w % 16)
            h = int((newW / w) * h)
            w = int(newW)
        if h % 2 > 0:
            h -= 1
        return w, h

    def _mjpegLink(self, node, xout, nodeOutput):
        print("Creating MJPEG link for {} node and {} xlink stream...".format(node.getName(), xout.getStreamName()))
        videnc = self.pipeline.createVideoEncoder()
        if isinstance(node, dai.node.ColorCamera):
            if node.video == nodeOutput:
                size = self.__calcEncodeableSize(node.getVideoSize())
                node.setVideoSize(size)
                videnc.setDefaultProfilePreset(node.getFps(), dai.VideoEncoderProperties.Profile.MJPEG)
            elif node.preview == nodeOutput:
                size = self.__calcEncodeableSize(node.getPreviewSize())
                node.setPreviewSize(size)
                videnc.setDefaultProfilePreset(node.getFps(), dai.VideoEncoderProperties.Profile.MJPEG)
            elif node.still == nodeOutput:
                size = self.__calcEncodeableSize(node.getStillSize())
                node.setStillSize(size)
                videnc.setDefaultProfilePreset(node.getFps(), dai.VideoEncoderProperties.Profile.MJPEG)

            nodeOutput.link(videnc.input)
        elif isinstance(node, dai.node.MonoCamera):
            videnc.setDefaultProfilePreset(node.getFps(), dai.VideoEncoderProperties.Profile.MJPEG)
            nodeOutput.link(videnc.input)
        elif isinstance(node, dai.node.StereoDepth):
            cameraNode = getattr(self.nodes, 'monoLeft', getattr(self.nodes, 'monoRight', None))
            if cameraNode is None:
                raise RuntimeError("Unable to find mono camera node to determine frame size!")
            videnc.setDefaultProfilePreset(cameraNode.getFps(), dai.VideoEncoderProperties.Profile.MJPEG)
            nodeOutput.link(videnc.input)
        elif isinstance(node, dai.NeuralNetwork):
            w, h = self.__calcEncodeableSize(self.nnManager.inputSize)
            manip = self.pipeline.createImageManip()
            manip.initialConfig.setResize(w, h)

            videnc.setDefaultProfilePreset(30, dai.VideoEncoderProperties.Profile.MJPEG)
            nodeOutput.link(manip.inputImage)
            manip.out.link(videnc.input)
        else:
            raise NotImplementedError("Unable to create mjpeg link for encountered node type: {}".format(type(node)))
        videnc.setQuality(self.poeQuality)
        videnc.bitstream.link(xout.input)

    def createColorCam(self, previewSize=None, res=dai.ColorCameraProperties.SensorResolution.THE_1080_P, fps=30, fullFov=True, orientation: dai.CameraImageOrientation=None, colorOrder=dai.ColorCameraProperties.ColorOrder.BGR, xout=False, xoutVideo=False, xoutStill=False):
        """
        Creates :obj:`depthai.node.ColorCamera` node based on specified attributes

        Args:
            previewSize (tuple, Optional): Size of the preview - :code:`(width, height)`
            res (depthai.ColorCameraProperties.SensorResolution, Optional): Camera resolution to be used
            fps (int, Optional): Camera FPS set on the device. Can limit / increase the amount of frames produced by the camera
            fullFov (bool, Optional): If set to :code:`True`, full frame will be scaled down to nn size. If to :code:`False`,
                it will first center crop the frame to meet the NN aspect ratio and then scale down the image.
            orientation (depthai.CameraImageOrientation, Optional): Custom camera orientation to be set on the device
            colorOrder (depthai.ColorCameraProperties, Optional): Color order to be used
            xout (bool, Optional): If set to :code:`True`, a dedicated :obj:`depthai.node.XLinkOut` will be created for this node
            xoutVideo (bool, Optional): If set to :code:`True`, a dedicated :obj:`depthai.node.XLinkOut` will be created for `video` output of this node
            xoutStill (bool, Optional): If set to :code:`True`, a dedicated :obj:`depthai.node.XLinkOut` will be created for `still` output of this node
        """
        self.nodes.camRgb = self.pipeline.createColorCamera()
        if previewSize is not None:
            self.nodes.camRgb.setPreviewSize(*previewSize)
        self.nodes.camRgb.setInterleaved(False)
        self.nodes.camRgb.setResolution(res)
        self.nodes.camRgb.setColorOrder(colorOrder)
        self.nodes.camRgb.setFps(fps)
        if orientation is not None:
            self.nodes.camRgb.setImageOrientation(orientation)
        self.nodes.camRgb.setPreviewKeepAspectRatio(not fullFov)

        if xout:
            self.nodes.xoutRgb = self.pipeline.createXLinkOut()
            self.nodes.xoutRgb.setStreamName(Previews.color.name)
            if self.lowBandwidth and not self.lowCapabilities:
                self._mjpegLink(self.nodes.camRgb, self.nodes.xoutRgb, self.nodes.camRgb.video)
            else:
                self.nodes.camRgb.preview.link(self.nodes.xoutRgb.input)
        if xoutVideo:
            self.nodes.xoutRgbVideo = self.pipeline.createXLinkOut()
            self.nodes.xoutRgbVideo.setStreamName(Previews.color.name + "_video")
            self.nodes.camRgb.video.link(self.nodes.xoutRgbVideo.input)
        if xoutStill:
            self.nodes.xoutRgbStill = self.pipeline.createXLinkOut()
            self.nodes.xoutRgbStill.setStreamName(Previews.color.name + "_still")
            self.nodes.camRgb.still.link(self.nodes.xoutRgbStill.input)
        self.nodes.xinRgbControl = self.pipeline.createXLinkIn()
        self.nodes.xinRgbControl.setMaxDataSize(1024)
        self.nodes.xinRgbControl.setStreamName(Previews.color.name + "_control")
        self.nodes.xinRgbControl.out.link(self.nodes.camRgb.inputControl)


    def createLeftCam(self, res=None, fps=30, orientation: dai.CameraImageOrientation=None, xout=False):
        """
        Creates :obj:`depthai.node.MonoCamera` node based on specified attributes, assigned to :obj:`depthai.CameraBoardSocket.LEFT`

        Args:
            res (depthai.MonoCameraProperties.SensorResolution, Optional): Camera resolution to be used
            fps (int, Optional): Camera FPS set on the device. Can limit / increase the amount of frames produced by the camera
            orientation (depthai.CameraImageOrientation, Optional): Custom camera orientation to be set on the device
            xout (bool, Optional): If set to :code:`True`, a dedicated :obj:`depthai.node.XLinkOut` will be created for this node
        """
        self.nodes.monoLeft = self.pipeline.createMonoCamera()
        self.nodes.monoLeft.setBoardSocket(dai.CameraBoardSocket.LEFT)
        if orientation is not None:
            self.nodes.monoLeft.setImageOrientation(orientation)
        if res is None:
            self.nodes.monoLeft.setResolution(dai.MonoCameraProperties.SensorResolution.THE_400_P if self.lowBandwidth and self.lowCapabilities else dai.MonoCameraProperties.SensorResolution.THE_720_P)
        else:
            self.nodes.monoLeft.setResolution(res)
        self.nodes.monoLeft.setFps(fps)

        self.nodes.xoutLeft = self.pipeline.createXLinkOut()
        self.nodes.xoutLeft.setStreamName(Previews.left.name)
        if xout:
            if self.lowBandwidth and not self.lowCapabilities:
                self._mjpegLink(self.nodes.monoLeft, self.nodes.xoutLeft, self.nodes.monoLeft.out)
            else:
                self.nodes.monoLeft.out.link(self.nodes.xoutLeft.input)
        self.nodes.xinLeftControl = self.pipeline.createXLinkIn()
        self.nodes.xinLeftControl.setMaxDataSize(1024)
        self.nodes.xinLeftControl.setStreamName(Previews.left.name + "_control")
        self.nodes.xinLeftControl.out.link(self.nodes.monoLeft.inputControl)

    def createRightCam(self, res=None, fps=30, orientation: dai.CameraImageOrientation=None, xout=False):
        """
        Creates :obj:`depthai.node.MonoCamera` node based on specified attributes, assigned to :obj:`depthai.CameraBoardSocket.RIGHT`

        Args:
            res (depthai.MonoCameraProperties.SensorResolution, Optional): Camera resolution to be used
            fps (int, Optional): Camera FPS set on the device. Can limit / increase the amount of frames produced by the camera
            orientation (depthai.CameraImageOrientation, Optional): Custom camera orientation to be set on the device
            xout (bool, Optional): If set to :code:`True`, a dedicated :obj:`depthai.node.XLinkOut` will be created for this node
        """
        self.nodes.monoRight = self.pipeline.createMonoCamera()
        self.nodes.monoRight.setBoardSocket(dai.CameraBoardSocket.RIGHT)
        if orientation is not None:
            self.nodes.monoRight.setImageOrientation(orientation)
        if res is None:
            self.nodes.monoRight.setResolution(dai.MonoCameraProperties.SensorResolution.THE_400_P if self.lowBandwidth and self.lowCapabilities else dai.MonoCameraProperties.SensorResolution.THE_720_P)
        else:
            self.nodes.monoRight.setResolution(res)
        self.nodes.monoRight.setFps(fps)

        self.nodes.xoutRight = self.pipeline.createXLinkOut()
        self.nodes.xoutRight.setStreamName(Previews.right.name)
        if xout:
            if self.lowBandwidth and not self.lowCapabilities:
                self._mjpegLink(self.nodes.monoRight, self.nodes.xoutRight, self.nodes.monoRight.out)
            else:
                self.nodes.monoRight.out.link(self.nodes.xoutRight.input)
        self.nodes.xinRightControl = self.pipeline.createXLinkIn()
        self.nodes.xinRightControl.setMaxDataSize(1024)
        self.nodes.xinRightControl.setStreamName(Previews.right.name + "_control")
        self.nodes.xinRightControl.out.link(self.nodes.monoRight.inputControl)

    def updateIrConfig(self, device, irLaser=None, irFlood=None):
        """
        Updates IR configuration

        Args:
            irLaser (int, Optional): Sets the IR laser dot projector brightness (0..1200)
            irFlood (int, Optional): Sets the IR flood illuminator light brightness (0..1500)
        """
        if irLaser is not None:
            device.setIrLaserDotProjectorBrightness(irLaser)
        if irFlood is not None:
            device.setIrFloodLightBrightness(irFlood)

    def createDepth(self, dct=245, median=dai.MedianFilter.KERNEL_7x7, sigma=0, lr=False, lrcThreshold=4, extended=False, subpixel=False, useDisparity=False, useDepth=False, useRectifiedLeft=False, useRectifiedRight=False):
        """
        Creates :obj:`depthai.node.StereoDepth` node based on specified attributes

        Args:
            dct (int, Optional): Disparity Confidence Threshold (0..255). The less confident the network is, the more empty values
                are present in the depth map.
            median (depthai.MedianFilter, Optional): Median filter to be applied on the depth, use with :obj:`depthai.MedianFilter.MEDIANOFF` to disable median filtering
            sigma (int, Optional): Sigma value for bilateral filter (0..65535). If set to :code:`0`, the filter will be disabled.
            lr (bool, Optional): Set to :code:`True` to enable Left-Right Check
            lrcThreshold (int, Optional): Sets the Left-Right Check threshold value (0..10)
            extended (bool, Optional): Set to :code:`True` to enable the extended disparity
            subpixel (bool, Optional): Set to :code:`True` to enable the subpixel disparity
            useDisparity (bool, Optional): Set to :code:`True` to create output queue for disparity frames
            useDepth (bool, Optional): Set to :code:`True` to create output queue for depth frames
            useRectifiedLeft (bool, Optional): Set to :code:`True` to create output queue for rectified left frames
            useRectifiedRigh (bool, Optional): Set to :code:`True` to create output queue for rectified right frames

        Raises:
            RuntimeError: if left of right mono cameras were not initialized
        """
        self.nodes.stereo = self.pipeline.createStereoDepth()

        self.nodes.stereo.initialConfig.setConfidenceThreshold(dct)
        self.nodes.stereo.initialConfig.setMedianFilter(median)
        self.nodes.stereo.initialConfig.setBilateralFilterSigma(sigma)
        self.nodes.stereo.initialConfig.setLeftRightCheckThreshold(lrcThreshold)

        # self.nodes.stereo.setRuntimeModeSwitch(True)
        self.nodes.stereo.setLeftRightCheck(lr)
        self.nodes.stereo.setExtendedDisparity(extended)
        self.nodes.stereo.setSubpixel(subpixel)

        self._depthConfig = self.nodes.stereo.initialConfig.get()

        # Create mono left/right cameras if we haven't already
        if not hasattr(self.nodes, 'monoLeft'):
            raise RuntimeError("Left mono camera not initialized. Call createLeftCam(res, fps) first!")
        if not hasattr(self.nodes, 'monoRight'):
            raise RuntimeError("Right mono camera not initialized. Call createRightCam(res, fps) first!")

        self.nodes.monoLeft.out.link(self.nodes.stereo.left)
        self.nodes.monoRight.out.link(self.nodes.stereo.right)

        self.nodes.xinStereoConfig = self.pipeline.createXLinkIn()
        self.nodes.xinStereoConfig.setMaxDataSize(1024)
        self.nodes.xinStereoConfig.setStreamName("stereoConfig")
        self.nodes.xinStereoConfig.out.link(self.nodes.stereo.inputConfig)

        if useDepth:
            self.nodes.xoutDepth = self.pipeline.createXLinkOut()
            self.nodes.xoutDepth.setStreamName(Previews.depthRaw.name)
            # if self.lowBandwidth and not self.lowCapabilities:  TODO change once depth frame type (14) is supported by VideoEncoder
            if False:
                self._mjpegLink(self.nodes.stereo, self.nodes.xoutDepth, self.nodes.stereo.depth)
            else:
                self.nodes.stereo.depth.link(self.nodes.xoutDepth.input)

        if useDisparity:
            self.nodes.xoutDisparity = self.pipeline.createXLinkOut()
            self.nodes.xoutDisparity.setStreamName(Previews.disparity.name)
            if False:
                self._mjpegLink(self.nodes.stereo, self.nodes.xoutDisparity, self.nodes.stereo.disparity)
            else:
                self.nodes.stereo.disparity.link(self.nodes.xoutDisparity.input)

        if useRectifiedLeft:
            self.nodes.xoutRectLeft = self.pipeline.createXLinkOut()
            self.nodes.xoutRectLeft.setStreamName(Previews.rectifiedLeft.name)
            # if self.lowBandwidth:  # disabled to limit the memory usage
            if False:
                self._mjpegLink(self.nodes.stereo, self.nodes.xoutRectLeft, self.nodes.stereo.rectifiedLeft)
            else:
                self.nodes.stereo.rectifiedLeft.link(self.nodes.xoutRectLeft.input)

        if useRectifiedRight:
            self.nodes.xoutRectRight = self.pipeline.createXLinkOut()
            self.nodes.xoutRectRight.setStreamName(Previews.rectifiedRight.name)
            # if self.lowBandwidth:  # disabled to limit the memory usage
            if False:
                self._mjpegLink(self.nodes.stereo, self.nodes.xoutRectRight, self.nodes.stereo.rectifiedRight)
            else:
                self.nodes.stereo.rectifiedRight.link(self.nodes.xoutRectRight.input)

    def _updateCamConfig(self, configRef: dai.CameraControl, exposure=None, sensitivity=None, saturation=None, contrast=None, brightness=None, sharpness=None):
        if any([exposure, sensitivity]):
            if not all([exposure, sensitivity]):
                raise RuntimeError("Both \"exposure\" and \"sensitivity\" arguments must be provided")
            configRef.setManualExposure(exposure, sensitivity)
        if saturation is not None:
            configRef.setSaturation(saturation)
        if sharpness is not None:
            configRef.setSharpness(sharpness)
        if contrast is not None:
            configRef.setContrast(contrast)
        if brightness is not None:
            configRef.setBrightness(brightness)

    def captureStill(self):
        ctrl = dai.CameraControl()
        ctrl.setCaptureStill(True)
        self._rgbConfigInputQueue.send(ctrl)

    def triggerAutoFocus(self):
        ctrl = dai.CameraControl()
        ctrl.setAutoFocusMode(dai.CameraControl.AutoFocusMode.AUTO)
        ctrl.setAutoFocusTrigger()
        self._rgbConfigInputQueue.send(ctrl)

    def triggerAutoExposure(self):
        ctrl = dai.CameraControl()
        ctrl.setAutoExposureEnable()
        self._rgbConfigInputQueue.send(ctrl)

    def triggerAutoWhiteBalance(self):
        ctrl = dai.CameraControl()
        ctrl.setAutoWhiteBalanceMode(dai.CameraControl.AutoWhiteBalanceMode.AUTO)
        self._rgbConfigInputQueue.send(ctrl)

    def updateColorCamConfig(self, exposure=None, sensitivity=None, saturation=None, contrast=None, brightness=None, sharpness=None, autofocus=None, autowhitebalance=None, focus=None, whitebalance=None):
        """
        Updates :obj:`depthai.node.ColorCamera` node config

        Args:
            exposure (int, Optional): Exposure time in microseconds. Has to be set together with :obj:`sensitivity` (Usual range: 1..33000)
            sensitivity (int, Optional): Sensivity as ISO value. Has to be set together with :obj:`exposure` (Usual range: 100..1600)
            saturation (int, Optional): Image saturation (Allowed range: -10..10)
            contrast (int, Optional): Image contrast (Allowed range: -10..10)
            brightness (int, Optional): Image brightness (Allowed range: -10..10)
            sharpness (int, Optional): Image sharpness (Allowed range: 0..4)
            autofocus (dai.CameraControl.AutoFocusMode, Optional): Set the autofocus mode
            autowhitebalance (dai.CameraControl.AutoFocusMode, Optional): Set the autowhitebalance mode
            focus (int, Optional): Set the manual focus (lens position)
            whitebalance (int, Optional): Set the manual white balance
        """
        self._updateCamConfig(self._rgbConfig, exposure, sensitivity, saturation, contrast, brightness, sharpness)
        if autofocus is not None:
            self._rgbConfig.setAutoFocusMode(autofocus)
        if autowhitebalance is not None:
            self._rgbConfig.setAutoWhiteBalanceMode(autowhitebalance)
        if focus is not None:
            self._rgbConfig.setManualFocus(focus)
        if whitebalance is not None:
            self._rgbConfig.setManualWhiteBalance(whitebalance)

        if any([exposure, sensitivity, saturation, contrast, brightness, sharpness, autofocus, autowhitebalance, focus, whitebalance]):
            self._rgbConfigInputQueue.send(self._rgbConfig)

    def updateLeftCamConfig(self, exposure=None, sensitivity=None, saturation=None, contrast=None, brightness=None, sharpness=None):
        """
        Updates left :obj:`depthai.node.MonoCamera` node config

        Args:
            exposure (int, Optional): Exposure time in microseconds. Has to be set together with :obj:`sensitivity` (Usual range: 1..33000)
            sensitivity (int, Optional): Sensivity as ISO value. Has to be set together with :obj:`exposure` (Usual range: 100..1600)
            saturation (int, Optional): Image saturation (Allowed range: -10..10)
            contrast (int, Optional): Image contrast (Allowed range: -10..10)
            brightness (int, Optional): Image brightness (Allowed range: -10..10)
            sharpness (int, Optional): Image sharpness (Allowed range: 0..4)
        """
        self._updateCamConfig(self._leftConfig, exposure, sensitivity, saturation, contrast, brightness, sharpness)
        self._leftConfigInputQueue.send(self._leftConfig)

    def updateRightCamConfig(self, exposure=None, sensitivity=None, saturation=None, contrast=None, brightness=None, sharpness=None):
        """
        Updates right :obj:`depthai.node.MonoCamera` node config

        Args:
            exposure (int, Optional): Exposure time in microseconds. Has to be set together with :obj:`sensitivity` (Usual range: 1..33000)
            sensitivity (int, Optional): Sensivity as ISO value. Has to be set together with :obj:`exposure` (Usual range: 100..1600)
            saturation (int, Optional): Image saturation (Allowed range: -10..10)
            contrast (int, Optional): Image contrast (Allowed range: -10..10)
            brightness (int, Optional): Image brightness (Allowed range: -10..10)
            sharpness (int, Optional): Image sharpness (Allowed range: 0..4)
        """
        self._updateCamConfig(self._rightConfig, exposure, sensitivity, saturation, contrast, brightness, sharpness)
        self._rightConfigInputQueue.send(self._rightConfig)

    def updateDepthConfig(self, dct=None, sigma=None, median=None, lrcThreshold=None):
        """
        Updates :obj:`depthai.node.StereoDepth` node config

        Args:
            dct (int, Optional): Disparity Confidence Threshold (0..255). The less confident the network is, the more empty values
                are present in the depth map.
            median (depthai.MedianFilter, Optional): Median filter to be applied on the depth, use with :obj:`depthai.MedianFilter.MEDIANOFF` to disable median filtering
            sigma (int, Optional): Sigma value for bilateral filter (0..65535). If set to :code:`0`, the filter will be disabled.
            lrc (bool, Optional): Enables or disables Left-Right Check mode
            lrcThreshold (int, Optional): Sets the Left-Right Check threshold value (0..10)
        """
        if any([dct, sigma, median, lrcThreshold]):
            if dct is not None:
                self._depthConfig.costMatching.confidenceThreshold = dct
            if sigma is not None:
                self._depthConfig.postProcessing.bilateralSigmaValue = sigma
            if median is not None:
                self._depthConfig.postProcessing.median = median
            if lrcThreshold is not None:
                self._depthConfig.algorithmControl.leftRightCheckThreshold = lrcThreshold
            self._depthConfigInputQueue.send(self._depthConfig)

    def addNn(self, nn, xoutNnInput=False, xoutSbb=False):
        """
        Adds NN node to current pipeline. Usually obtained by calling :obj:`depthai_sdk.managers.NNetManager.createNN` method
        first

        Args:
            nn (depthai.node.NeuralNetwork): prepared NeuralNetwork node to be attached to the pipeline
            xoutNnInput (bool): Set to :code:`True` to create output queue for NN's passthough frames
            xoutSbb (bool): Set to :code:`True` to create output queue for Spatial Bounding Boxes (area that is used to calculate spatial location)
        """
        # TODO adjust this function once passthrough frame type (8) is supported by VideoEncoder (for self.MjpegLink)
        if xoutNnInput:
            self.nodes.xoutNnInput = self.pipeline.createXLinkOut()
            self.nodes.xoutNnInput.setStreamName(Previews.nnInput.name)
            nn.passthrough.link(self.nodes.xoutNnInput.input)

        if xoutSbb and self.nnManager._nnFamily in ("YOLO", "mobilenet"):
            self.nodes.xoutSbb = self.pipeline.createXLinkOut()
            self.nodes.xoutSbb.setStreamName("sbb")
            nn.boundingBoxMapping.link(self.nodes.xoutSbb.input)

    def createSystemLogger(self, rate=1):
        """
        Creates :obj:`depthai.node.SystemLogger` node together with XLinkOut

        Args:
            rate (int, Optional): Specify logging rate (in Hz)
        """
        self.nodes.systemLogger = self.pipeline.createSystemLogger()
        self.nodes.systemLogger.setRate(rate)
        self.nodes.xoutSystemLogger = self.pipeline.createXLinkOut()
        self.nodes.xoutSystemLogger.setStreamName("systemLogger")
        self.nodes.systemLogger.out.link(self.nodes.xoutSystemLogger.input)

    def createEncoder(self, cameraName, encFps=30, encQuality=100):
        """
        Creates H.264 / H.265 video encoder (:obj:`depthai.node.VideoEncoder` instance)

        Args:
            cameraName (str): Camera name to create the encoder for
            encFps (int, Optional): Specify encoding FPS
            encQuality (int, Optional): Specify encoding quality (1-100)

        Raises:
            ValueError: if cameraName is not a supported camera name
            RuntimeError: if specified camera node was not present
        """
        allowedSources = [Previews.left.name, Previews.right.name, Previews.color.name]
        if cameraName not in allowedSources:
            raise ValueError(
                "Camera param invalid, received {}, available choices: {}".format(cameraName, allowedSources))
        nodeName = cameraName.lower() + 'Enc'
        xoutName = nodeName + "Xout"
        encProfile = dai.VideoEncoderProperties.Profile.H264_MAIN

        if cameraName == Previews.color.name:
            if not hasattr(self.nodes, 'camRgb'):
                raise RuntimeError("RGB camera not initialized. Call createColorCam(res, fps) first!")
            encProfile = dai.VideoEncoderProperties.Profile.H264_MAIN
            encIn = self.nodes.camRgb.video

        elif cameraName == Previews.left.name:
            if not hasattr(self.nodes, 'monoLeft'):
                raise RuntimeError("Left mono camera not initialized. Call createLeftCam(res, fps) first!")
            encIn = self.nodes.monoLeft.out
        elif cameraName == Previews.right.name:
            if not hasattr(self.nodes, 'monoRight'):
                raise RuntimeError("Right mono camera not initialized. Call createRightCam(res, fps) first!")
            encIn = self.nodes.monoRight.out

        enc = self.pipeline.createVideoEncoder()
        enc.setDefaultProfilePreset(encFps, encProfile)
        enc.setQuality(encQuality)
        encIn.link(enc.input)
        setattr(self.nodes, nodeName, enc)

        encXout = self.pipeline.createXLinkOut()
        enc.bitstream.link(encXout.input)
        encXout.setStreamName(xoutName)
        setattr(self.nodes, xoutName, encXout)

    def enableLowBandwidth(self, poeQuality):
        """
        Enables low-bandwidth mode

        Args:
            poeQuality (int, Optional): PoE encoding quality, can decrease frame quality but decrease latency
        """
        self.lowBandwidth = True
        self.poeQuality = poeQuality

    def setXlinkChunkSize(self, chunkSize):
        self.pipeline.setXLinkChunkSize(chunkSize)



import json
from pathlib import Path
import depthai as dai
import cv2
import numpy as np

from .preview_manager import PreviewManager, SyncedPreviewManager
from ..previews import Previews
from ..utils import loadModule, toTensorResult, frameNorm, toPlanar


class NNetManager:
    """
    Manager class handling all NN-related functionalities. It's capable of creating appropriate nodes and connections,
    decoding neural network output automatically or by using external handler file.
    """

    def __init__(self, inputSize, nnFamily=None, labels=[], confidence=0.5, sync=False):
        """
        Args:
            inputSize (tuple): Desired NN input size, should match the input size defined in the network itself (width, height)
            nnFamily (str, Optional): type of NeuralNetwork to be processed. Supported: :code:`"YOLO"` and :code:`mobilenet`
            labels (list, Optional): Allows to display class label instead of ID when drawing nn detections.
            confidence (float, Optional): Specify detection nn's confidence threshold
            sync (bool, Optional): Store NN results for preview syncing (to be used with SyncedPreviewManager
        """
        self.inputSize = inputSize
        self._nnFamily = nnFamily
        if nnFamily in ("YOLO", "mobilenet"):
            self._outputFormat = "detection"
        self._labels = labels
        self._confidence = confidence
        self._sync = sync

    #: list: List of available neural network inputs
    sourceChoices = ("color", "left", "right", "rectifiedLeft", "rectifiedRight", "host")
    #: str: Selected neural network input
    source = None
    #: tuple: NN input size (width, height)
    inputSize = None
    #: depthai.OpenVINO.Version: OpenVINO version, available only if parsed from config file (see :func:`readConfig`)
    openvinoVersion = None
    #: depthai.DataInputQueue: DepthAI input queue object that allows to send images from host to device (used only with :code:`host` source)
    inputQueue = None
    #: depthai.DataOutputQueue: DepthAI output queue object that allows to receive NN results from the device.
    outputQueue = None
    #: dict: nn data buffer, disabled by default. Stores parsed nn data with packet sequence number as dict key
    buffer = {}


    _bboxColors = np.random.random(size=(256, 3)) * 256  # Random Colors for bounding boxes
    _countLabel = None
    _textBgColor = (0, 0, 0)
    _textColor = (255, 255, 255)
    _lineType = cv2.LINE_AA
    _textType = cv2.FONT_HERSHEY_SIMPLEX
    _outputFormat = "raw"
    _metadata = None
    _fullFov = True
    _config = None
    _nnFamily = None
    _handler = None

    def readConfig(self, path):
        """
        Parses the model config file and adjusts NNetManager values accordingly. It's advised to create a config file
        for every new network, as it allows to use dedicated NN nodes (for `MobilenetSSD <https://github.com/luxonis/depthai/blob/main/resources/nn/mobilenet-ssd/mobilenet-ssd.json>`__ and `YOLO <https://github.com/luxonis/depthai/blob/main/resources/nn/tiny-yolo-v3/tiny-yolo-v3.json>`__)
        or use `custom handler <https://github.com/luxonis/depthai/blob/main/resources/nn/openpose2/openpose2.json>`__ to process and display custom network results

        Args:
            path (pathlib.Path): Path to model config file (.json)

        Raises:
            ValueError: If path to config file does not exist
            RuntimeError: If custom handler does not contain :code:`draw` or :code:`show` methods
        """
        configPath = Path(path)
        if not configPath.exists():
            raise ValueError("Path {} does not exist!".format(path))

        with configPath.open() as f:
            self._config = json.load(f)
            if "openvino_version" in self._config:
                self.openvinoVersion =getattr(dai.OpenVINO.Version, 'VERSION_' + self._config.get("openvino_version"))
            nnConfig = self._config.get("nn_config", {})
            self._labels = self._config.get("mappings", {}).get("labels", None)
            self._nnFamily = nnConfig.get("NN_family", None)
            self._outputFormat = nnConfig.get("output_format", "raw")
            self._metadata = nnConfig.get("NN_specific_metadata", {})
            if "input_size" in nnConfig:
                self.inputSize = tuple(map(int, nnConfig.get("input_size").split('x')))

            self._confidence = self._metadata.get("confidence_threshold", nnConfig.get("confidence_threshold", None))
            if 'handler' in self._config:
                self._handler = loadModule(configPath.parent / self._config["handler"])

                if not callable(getattr(self._handler, "draw", None)) or not callable(getattr(self._handler, "decode", None)):
                    raise RuntimeError("Custom model handler does not contain 'draw' or 'decode' methods!")


    def _normFrame(self, frame):
        if not self._fullFov:
            scaleF = frame.shape[0] / self.inputSize[1]
            return np.zeros((int(self.inputSize[1] * scaleF), int(self.inputSize[0] * scaleF)))
        else:
            return frame

    def _cropOffsetX(self, frame):
        if not self._fullFov:
            croppedW = (frame.shape[0] / self.inputSize[1]) * self.inputSize[0]
            return int((frame.shape[1] - croppedW) // 2)
        else:
            return 0

    def createNN(self, pipeline, nodes, blobPath, source="color", useDepth=False, minDepth=100, maxDepth=10000, sbbScaleFactor=0.3, fullFov=True):
        """
        Creates nodes and connections in provided pipeline that will allow to run NN model and consume it's results.

        Args:
            pipeline (depthai.Pipeline): Pipeline instance
            nodes (types.SimpleNamespace): Object cointaining all of the nodes added to the pipeline. Available in :attr:`depthai_sdk.managers.PipelineManager.nodes`
            blobPath (pathlib.Path): Path to MyriadX blob. Might be useful to use together with
                :func:`depthai_sdk.managers.BlobManager.getBlob()` for dynamic blob compilation
            source (str, Optional): Neural network input source, one of :attr:`sourceChoices`
            useDepth (bool, Optional): If set to True, produced detections will have spatial coordinates included
            minDepth (int, Optional): Minimum depth distance in centimeters
            maxDepth (int, Optional): Maximum depth distance in centimeters
            sbbScaleFactor (float, Optional): Scale of the bounding box that will be used to calculate spatial coordinates for
                detection. If set to 0.3, it will scale down center-wise the bounding box to 0.3 of it's original size
                and use it to calculate spatial location of the object
            fullFov (bool, Optional): If set to False, manager will include crop offset when scaling the detections.
                Usually should be set to True (if you don't perform aspect ratio crop or when `keepAspectRatio` flag
                on camera/manip node is set to False

        Returns:
            depthai.node.NeuralNetwork: Configured NN node that was added to the pipeline

        Raises:
            RuntimeError: If source is not a valid choice or when input size has not been set.
        """
        if source not in self.sourceChoices:
            raise RuntimeError(f"Source {source} is invalid, available {self.sourceChoices}")
        if self.inputSize is None:
            raise RuntimeError("Unable to determine the nn input size. Please use --cnnInputSize flag to specify it in WxH format: -nnSize <width>x<height>")

        self.source = source
        self._fullFov = fullFov
        if self._nnFamily == "mobilenet":
            nodes.nn = pipeline.createMobileNetSpatialDetectionNetwork() if useDepth else pipeline.createMobileNetDetectionNetwork()
            nodes.nn.setConfidenceThreshold(self._confidence)
        elif self._nnFamily == "YOLO":
            nodes.nn = pipeline.createYoloSpatialDetectionNetwork() if useDepth else pipeline.createYoloDetectionNetwork()
            nodes.nn.setConfidenceThreshold(self._confidence)
            nodes.nn.setNumClasses(self._metadata["classes"])
            nodes.nn.setCoordinateSize(self._metadata["coordinates"])
            nodes.nn.setAnchors(self._metadata["anchors"])
            nodes.nn.setAnchorMasks(self._metadata["anchor_masks"])
            nodes.nn.setIouThreshold(self._metadata["iou_threshold"])
        else:
            # TODO use createSpatialLocationCalculator
            nodes.nn = pipeline.createNeuralNetwork()

        nodes.nn.setBlobPath(str(blobPath))
        nodes.nn.setNumInferenceThreads(2)
        nodes.nn.input.setBlocking(False)
        nodes.nn.input.setQueueSize(2)

        nodes.xoutNn = pipeline.createXLinkOut()
        nodes.xoutNn.setStreamName("nnOut")
        nodes.nn.out.link(nodes.xoutNn.input)

        if self.source == "host":
            nodes.xinNn = pipeline.createXLinkIn()
            nodes.xinNn.setMaxDataSize(self.inputSize[0] * self.inputSize[1] * 3)
            nodes.xinNn.setStreamName("nnIn")
            nodes.xinNn.out.link(nodes.nn.input)
        else:
            nodes.manipNn = pipeline.createImageManip()
            nodes.manipNn.initialConfig.setResize(*self.inputSize)
            # The NN model expects BGR input. By default ImageManip output type would be same as input (gray in this case)
            nodes.manipNn.initialConfig.setFrameType(dai.RawImgFrame.Type.BGR888p)
            # NN inputs
            nodes.manipNn.out.link(nodes.nn.input)
            nodes.manipNn.setKeepAspectRatio(not self._fullFov)
            nodes.manipNn.setMaxOutputFrameSize(self.inputSize[0] * self.inputSize[1] * 3)

            if self.source == "color":
                nodes.camRgb.preview.link(nodes.manipNn.inputImage)
            if self.source == "left":
                nodes.monoLeft.out.link(nodes.manipNn.inputImage)
            elif self.source == "right":
                nodes.monoRight.out.link(nodes.manipNn.inputImage)
            elif self.source == "rectifiedLeft":
                nodes.stereo.rectifiedLeft.link(nodes.manipNn.inputImage)
            elif self.source == "rectifiedRight":
                nodes.stereo.rectifiedRight.link(nodes.manipNn.inputImage)

        if self._nnFamily in ("YOLO", "mobilenet") and useDepth:
            nodes.stereo.depth.link(nodes.nn.inputDepth)
            nodes.nn.setDepthLowerThreshold(minDepth)
            nodes.nn.setDepthUpperThreshold(maxDepth)
            nodes.nn.setBoundingBoxScaleFactor(sbbScaleFactor)

        return nodes.nn

    def getLabelText(self, label):
        """
        Retrieves text assigned to specific label

        Args:
            label (int): Integer representing detection label, usually returned from NN node

        Returns:
            str: Label text assigned to specific label id or label id

        Raises:
            RuntimeError: If source is not a valid choice or when input size has not been set.
        """
        if self._labels is None:
            return str(label)
        elif int(label) < len(self._labels):
            return self._labels[int(label)]
        else:
            print(f"Label of ouf bounds (label index: {label}, available labels: {len(self._labels)}")
            return str(label)

    def parse(self, blocking=False):
        if self.outputQueue is None:
            return None, None

        if blocking:
            inNn = self.outputQueue.get()
        else:
            inNn = self.outputQueue.tryGet()
        if inNn is not None:
            data = self.decode(inNn)
            if self._sync:
                self.buffer[inNn.getSequenceNum()] = data
            return data, inNn
        else:
            return None, None

    def decode(self, inNn):
        """
        Decodes NN output. Performs generic handling for supported detection networks or calls custom handler methods

        Args:
            inNn (depthai.NNData): Integer representing detection label, usually returned from NN node

        Returns:
            Decoded NN data

        Raises:
            RuntimeError: if outputFormat specified in model config file is not recognized
        """
        if self._outputFormat == "detection":
            return inNn.detections
        elif self._outputFormat == "raw":
            if self._handler is not None:
                return self._handler.decode(self, inNn)
            else:
                try:
                    data = toTensorResult(inNn)
                    print("Received NN packet: ", ", ".join([f"{key}: {value.shape}" for key, value in data.items()]))
                except Exception as ex:
                    print("Received NN packet: <Preview unabailable: {}>".format(ex))
        else:
            raise RuntimeError("Unknown output format: {}".format(self._outputFormat))

    def _drawCount(self, source, decodedData):
        def drawCnt(frame, cnt):
            cv2.putText(frame, f"{self._countLabel}: {cnt}", (5, 46), self._textType, 0.5, self._textBgColor, 4, self._lineType)
            cv2.putText(frame, f"{self._countLabel}: {cnt}", (5, 46), self._textType, 0.5, self._textColor, 1, self._lineType)

        # Count the number of detected objects
        cntList = list(filter(lambda x: self.getLabelText(x.label) == self._countLabel, decodedData))
        if isinstance(source, PreviewManager):
            for frame in source.frames.values():
                drawCnt(frame, len(cntList))
        else:
            drawCnt(source, len(cntList))

    def draw(self, source, decodedData):
        """
        Draws NN results onto the frames. It's responsible to correctly map the results onto each frame requested,
        including applying crop offset or preparing a correct normalization frame, then draws them with all information
        provided (confidence, label, spatial location, label count).

        Also, it's able to call custom nn handler method :code:`draw` to hand over drawing the results

        Args:
            source (depthai_sdk.managers.PreviewManager | numpy.ndarray): Draw target.
                If supplied with a regular frame, it will draw the count on that frame

                If supplied with :class:`depthai_sdk.managers.PreviewManager` instance, it will print the count label
                on all of the frames that it stores

            decodedData: Detections from neural network node, usually returned from :func:`decode` method
        """
        if self._outputFormat == "detection":
            def drawDetection(frame, detection):
                bbox = frameNorm(self._normFrame(frame), [detection.xmin, detection.ymin, detection.xmax, detection.ymax])
                if self.source == Previews.color.name and not self._fullFov:
                    bbox[::2] += self._cropOffsetX(frame)
                cv2.rectangle(frame, (bbox[0], bbox[1]), (bbox[2], bbox[3]), self._bboxColors[detection.label], 2)
                cv2.rectangle(frame, (bbox[0], (bbox[1] - 28)), ((bbox[0] + 110), bbox[1]), self._bboxColors[detection.label], cv2.FILLED)
                cv2.putText(frame, self.getLabelText(detection.label), (bbox[0] + 5, bbox[1] - 10),
                            self._textType, 0.5, (0, 0, 0), 1, self._lineType)
                cv2.putText(frame, f"{int(detection.confidence * 100)}%", (bbox[0] + 62, bbox[1] - 10),
                            self._textType, 0.5, (0, 0, 0), 1, self._lineType)

                if hasattr(detection, 'spatialCoordinates'):  # Display spatial coordinates as well
                    xMeters = detection.spatialCoordinates.x / 1000
                    yMeters = detection.spatialCoordinates.y / 1000
                    zMeters = detection.spatialCoordinates.z / 1000
                    cv2.putText(frame, "X: {:.2f} m".format(xMeters), (bbox[0] + 10, bbox[1] + 60),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "X: {:.2f} m".format(xMeters), (bbox[0] + 10, bbox[1] + 60),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
                    cv2.putText(frame, "Y: {:.2f} m".format(yMeters), (bbox[0] + 10, bbox[1] + 75),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "Y: {:.2f} m".format(yMeters), (bbox[0] + 10, bbox[1] + 75),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
                    cv2.putText(frame, "Z: {:.2f} m".format(zMeters), (bbox[0] + 10, bbox[1] + 90),
                                self._textType, 0.5, self._textBgColor, 4, self._lineType)
                    cv2.putText(frame, "Z: {:.2f} m".format(zMeters), (bbox[0] + 10, bbox[1] + 90),
                                self._textType, 0.5, self._textColor, 1, self._lineType)
            if isinstance(source, SyncedPreviewManager):
                if len(self.buffer) > 0 and source.nnSyncSeq is not None:
                    data = self.buffer.get(source.nnSyncSeq, self.buffer[max(self.buffer.keys())])
                    for old_key in list(filter(lambda key: key < source.nnSyncSeq, self.buffer.keys())):
                        del self.buffer[old_key]
                    for detection in data:
                        for name, frame in source.frames.items():
                            drawDetection(frame, detection)
            else:
                for detection in decodedData:
                    if isinstance(source, PreviewManager):
                        for name, frame in source.frames.items():
                            drawDetection(frame, detection)
                    else:
                        drawDetection(source, detection)

            if self._countLabel is not None:
                self._drawCount(source, decodedData)

        elif self._outputFormat == "raw" and self._handler is not None:
            if isinstance(source, PreviewManager):
                frames = list(source.frames.items())
            else:
                frames = [("host", source)]
            self._handler.draw(self, decodedData, frames)

    def createQueues(self, device):
        """
        Creates output queue for NeuralNetwork node and, if using :code:`host` as a :attr:`source`, it will also create
        input queue.

        Args:
            device (depthai.Device): Running device instance
        """
        if self.source == "host":
            self.inputQueue = device.getInputQueue("nnIn", maxSize=1, blocking=False)
        self.outputQueue = device.getOutputQueue("nnOut", maxSize=1, blocking=False)

    def closeQueues(self):
        """
        Closes output queues created by :func:`createQueues`
        """
        if self.source == "host" and self.inputQueue is not None:
            self.inputQueue.close()
        if self.outputQueue is not None:
            self.outputQueue.close()

    def sendInputFrame(self, frame, seqNum=None):
        """
        Sends a frame into :attr:`inputQueue` object. Handles scaling down the frame, creating a proper :obj:`depthai.ImgFrame`
        and sending it to the queue. Be sure to use :code:`host` as a :attr:`source` and call :func:`createQueues` prior
        input queue.

        Args:
            frame (numpy.ndarray): Frame to be sent to the device
            seqNum (int, Optional): Sequence number set on ImgFrame. Useful in synchronization scenarios

        Returns:
            numpy.ndarray: scaled frame that was sent to the NN (same width/height as NN input)

        Raises:
            RuntimeError: if :attr:`inputQueue` is :code:`None` (unable to send the image)
        """
        if self.inputQueue is None:
            raise RuntimeError("Unable to send image, no input queue is present! Call `createQueues(device)` first!")

        scaledFrame = cv2.resize(frame, self.inputSize)
        frameNn = dai.ImgFrame()
        if seqNum is not None:
            frameNn.setSequenceNum(seqNum)
        frameNn.setType(dai.ImgFrame.Type.BGR888p)
        frameNn.setWidth(self.inputSize[0])
        frameNn.setHeight(self.inputSize[1])
        frameNn.setData(toPlanar(scaledFrame))
        self.inputQueue.send(frameNn)

        return scaledFrame

    def countLabel(self, label):
        """
        Enables object count for specific label. Label count will be printed once :func:`draw` method is called

        Args:
            label (str | int): Label to be counted. If model is using mappings in model config file, supply here a :obj:`str` label
                to be tracked. If no mapping is present, specify the label as :obj:`int` (NN-default)
        """

        self._countLabel = label


import enum
import math
from functools import partial

import cv2
import numpy as np
try:
    from turbojpeg import TurboJPEG, TJFLAG_FASTUPSAMPLE, TJFLAG_FASTDCT, TJPF_GRAY
    turbo = TurboJPEG()
except:
    turbo = None


class PreviewDecoder:

    @staticmethod
    def __jpegDecode(data, type):
        if turbo is not None:
            if type == cv2.IMREAD_GRAYSCALE:
                return turbo.decode(data, flags=TJFLAG_FASTUPSAMPLE | TJFLAG_FASTDCT, pixel_format=TJPF_GRAY)
            if type == cv2.IMREAD_UNCHANGED:
                return turbo.decode_to_yuv(data, flags=TJFLAG_FASTUPSAMPLE | TJFLAG_FASTDCT)
            else:
                return turbo.decode(data, flags=TJFLAG_FASTUPSAMPLE | TJFLAG_FASTDCT)
        else:
            return cv2.imdecode(data, type)

    @staticmethod
    def nnInput(packet, manager=None):
        """
        Produces NN passthough frame from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        # if manager is not None and manager.decode: TODO change once passthrough frame type (8) is supported by VideoEncoder
        if False:
            frame = PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_COLOR)
        else:
            frame = packet.getCvFrame()
        if hasattr(manager, "nnSource") and manager.nnSource in (Previews.rectifiedLeft.name, Previews.rectifiedRight.name):
            frame = cv2.flip(frame, 1)
        return frame

    @staticmethod
    def color(packet, manager=None):
        """
        Produces color camera frame from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        if manager is not None and manager.decode:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_COLOR)
        else:
            return packet.getCvFrame()

    @staticmethod
    def left(packet, manager=None):
        """
        Produces left camera frame from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        if manager is not None and manager.decode:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
        else:
            return packet.getCvFrame()

    @staticmethod
    def right(packet, manager=None):
        """
        Produces right camera frame from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        if manager is not None and manager.decode:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
        else:
            return packet.getCvFrame()

    @staticmethod
    def rectifiedLeft(packet, manager=None):
        """
        Produces rectified left frame (:obj:`depthai.node.StereoDepth.rectifiedLeft`) from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        # if manager is not None and manager.decode:  # disabled to limit the memory usage
        if False:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
        else:
            return packet.getCvFrame()

    @staticmethod
    def rectifiedRight(packet, manager=None):
        """
        Produces rectified right frame (:obj:`depthai.node.StereoDepth.rectifiedRight`) from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        # if manager is not None and manager.decode:  # disabled to limit the memory usage
        if False:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
        else:
            return packet.getCvFrame()

    @staticmethod
    def depthRaw(packet, manager=None):
        """
        Produces raw depth frame (:obj:`depthai.node.StereoDepth.depth`) from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        # if manager is not None and manager.decode:  TODO change once depth frame type (14) is supported by VideoEncoder
        if False:
            return PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_UNCHANGED)
        else:
            return packet.getFrame()

    @staticmethod
    def depth(depthRaw, manager=None):
        """
        Produces depth frame from raw depth frame (converts to disparity and applies color map)

        Args:
            depthRaw (numpy.ndarray): OpenCV frame containing raw depth frame
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        if getattr(manager, "_depthConfig", None) is None:
            raise RuntimeError("Depth config has to be provided before decoding depth data")

        maxDisp = manager._depthConfig.getMaxDisparity()
        subpixelLevels = pow(2, manager._depthConfig.get().algorithmControl.subpixelFractionalBits)
        subpixel = manager._depthConfig.get().algorithmControl.enableSubpixel
        dispIntegerLevels = maxDisp if not subpixel else maxDisp / subpixelLevels
        dispScaleFactor = getattr(manager, "dispScaleFactor", None)
        if dispScaleFactor is None:
            baseline = getattr(manager, 'baseline', 75)  # mm
            fov = getattr(manager, 'fov', 71.86)
            focal = getattr(manager, 'focal', depthRaw.shape[1] / (2. * math.tan(math.radians(fov / 2))))
            dispScaleFactor = baseline * focal
            if manager is not None:
                setattr(manager, "dispScaleFactor", dispScaleFactor)
        with np.errstate(divide='ignore'):
            dispFrame = dispScaleFactor / depthRaw

        dispFrame = (dispFrame * 255. / dispIntegerLevels).astype(np.uint8)

        return PreviewDecoder.disparityColor(dispFrame, manager)

    @staticmethod
    def disparity(packet, manager=None):
        """
        Produces disparity frame (:obj:`depthai.node.StereoDepth.disparity`) from raw data packet

        Args:
            packet (depthai.ImgFrame): Packet received from output queue
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        if False:
            rawFrame = PreviewDecoder.__jpegDecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
        else:
            rawFrame = packet.getFrame()
        return (rawFrame*(manager.dispMultiplier if manager is not None else 255/96)).astype(np.uint8)

    @staticmethod
    def disparityColor(disparity, manager=None):
        """
        Applies color map to disparity frame

        Args:
            disparity (numpy.ndarray): OpenCV frame containing disparity frame
            manager (depthai_sdk.managers.PreviewManager, optional): PreviewManager instance

        Returns:
            numpy.ndarray: Ready to use OpenCV frame
        """
        return cv2.applyColorMap(disparity, manager.colorMap if manager is not None else cv2.COLORMAP_JET)


class Previews(enum.Enum):
    """
    Enum class, assigning preview name with decode function.

    Usually used as e.g. :code:`Previews.color.name` when specifying color preview name.

    Can be also used as e.g. :code:`Previews.color.value(packet)` to transform queue output packet to color camera frame
    """
    nnInput = partial(PreviewDecoder.nnInput)
    color = partial(PreviewDecoder.color)
    left = partial(PreviewDecoder.left)
    right = partial(PreviewDecoder.right)
    rectifiedLeft = partial(PreviewDecoder.rectifiedLeft)
    rectifiedRight = partial(PreviewDecoder.rectifiedRight)
    depthRaw = partial(PreviewDecoder.depthRaw)
    depth = partial(PreviewDecoder.depth)
    disparity = partial(PreviewDecoder.disparity)
    disparityColor = partial(PreviewDecoder.disparityColor)


class MouseClickTracker:
    """
    Class that allows to track the click events on preview windows and show pixel value of a frame in coordinates pointed
    by the user.

    Used internally by :obj:`depthai_sdk.managers.PreviewManager`
    """

    #: dict: Stores selected point position per frame
    points = {}
    #: dict: Stores values assigned to specific point per frame
    values = {}

    def selectPoint(self, name):
        """
        Returns callback function for :code:`cv2.setMouseCallback` that will update the selected point on mouse click
        event from frame.

        Usually used as

        .. code-block:: python

            mct = MouseClickTracker()
            # create preview window
            cv2.setMouseCallback(window_name, mct.select_point(window_name))

        Args:
            name (str): Name of the frame

        Returns:
            Callback function for :code:`cv2.setMouseCallback`
        """
        def cb(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONUP:
                if self.points.get(name) == (x, y):
                    del self.points[name]
                    if name in self.values:
                        del self.values[name]
                else:
                    self.points[name] = (x, y)
        return cb

    def extractValue(self, name, frame: np.ndarray):
        """
        Extracts value from frame for a specific point

        Args:
            name (str): Name of the frame
        """
        point = self.points.get(name, None)
        if point is not None and frame is not None:
            if name in (Previews.depthRaw.name, Previews.depth.name):
                self.values[name] = "{}mm".format(frame[point[1]][point[0]])
            elif name in (Previews.disparityColor.name, Previews.disparity.name):
                self.values[name] = "{}px".format(frame[point[1]][point[0]])
            elif len(frame.shape) == 3:
                self.values[name] = "R:{},G:{},B:{}".format(*frame[point[1]][point[0]][::-1])
            elif len(frame.shape) == 2:
                self.values[name] = "Gray:{}".format(frame[point[1]][point[0]])
            else:
                self.values[name] = str(frame[point[1]][point[0]])


import importlib
import sys
from pathlib import Path
import urllib.request

import cv2
import numpy as np
import depthai as dai
import datetime as dt
from heapq import heappop, heappush
import threading


def cosDist(a, b):
    """
    Calculates cosine distance - https://en.wikipedia.org/wiki/Cosine_similarity
    """
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def frameNorm(frame, bbox):
    """
    Mapps bounding box coordinates (0..1) to pixel values on frame

    Args:
        frame (numpy.ndarray): Frame to which adjust the bounding box
        bbox (list): list of bounding box points in a form of :code:`[x1, y1, x2, y2, ...]`

    Returns:
        list: Bounding box points mapped to pixel values on frame
    """
    normVals = np.full(len(bbox), frame.shape[0])
    normVals[::2] = frame.shape[1]
    return (np.clip(np.array(bbox), 0, 1) * normVals).astype(int)


def toPlanar(arr: np.ndarray, shape: tuple = None) -> np.ndarray:
    """
    Converts interleaved frame into planar

    Args:
        arr (numpy.ndarray): Interleaved frame
        shape (tuple, optional): If provided, the interleaved frame will be scaled to specified shape before converting into planar

    Returns:
        numpy.ndarray: Planar frame
    """
    if shape is None:
        return arr.transpose(2, 0, 1)
    return cv2.resize(arr, shape).transpose(2, 0, 1)


def toTensorResult(packet):
    """
    Converts NN packet to dict, with each key being output tensor name and each value being correctly reshaped and converted results array

    Useful as a first step of processing NN results for custom neural networks

    Args:
        packet (depthai.NNData): Packet returned from NN node

    Returns:
        dict: Dict containing prepared output tensors
    """
    data = {}
    for tensor in packet.getRaw().tensors:
        if tensor.dataType == dai.TensorInfo.DataType.INT:
            data[tensor.name] = np.array(packet.getLayerInt32(tensor.name)).reshape(tensor.dims)
        elif tensor.dataType == dai.TensorInfo.DataType.FP16:
            data[tensor.name] = np.array(packet.getLayerFp16(tensor.name)).reshape(tensor.dims)
        elif tensor.dataType == dai.TensorInfo.DataType.I8:
            data[tensor.name] = np.array(packet.getLayerUInt8(tensor.name)).reshape(tensor.dims)
        else:
            print("Unsupported tensor layer type: {}".format(tensor.dataType))
    return data


def merge(source:dict, destination:dict):
    """
    Utility function to merge two dictionaries

    .. code-block:: python

        a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        print(merge(b, a))
        # { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }

    Args:
        source (dict): first dict to merge
        destination (dict): second dict to merge

    Returns:
        dict: merged dict
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            merge(value, node)
        else:
            destination[key] = value

    return destination


def loadModule(path: Path):
    """
    Loads module from specified path. Used internally e.g. to load a custom handler file from path

    Args:
        path (pathlib.Path): path to the module to be loaded

    Returns:
        module: loaded module from provided path
    """
    spec = importlib.util.spec_from_file_location(path.stem, str(path.absolute()))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def getDeviceInfo(deviceId=None):
    """
    Find a correct :obj:`depthai.DeviceInfo` object, either matching provided :code:`deviceId` or selected by the user (if multiple devices available)
    Useful for almost every app where there is a possibility of multiple devices being connected simultaneously

    Args:
        deviceId (str, optional): Specifies device MX ID, for which the device info will be collected

    Returns:
        depthai.DeviceInfo: Object representing selected device info

    Raises:
        RuntimeError: if no DepthAI device was found or, if :code:`deviceId` was specified, no device with matching MX ID was found
        ValueError: if value supplied by the user when choosing the DepthAI device was incorrect
    """
    deviceInfos = dai.Device.getAllAvailableDevices()
    if len(deviceInfos) == 0:
        raise RuntimeError("No DepthAI device found!")
    else:
        print("Available devices:")
        for i, deviceInfo in enumerate(deviceInfos):
            print(f"[{i}] {deviceInfo.getMxId()} [{deviceInfo.state.name}]")

        if deviceId == "list":
            raise SystemExit(0)
        elif deviceId is not None:
            matchingDevice = next(filter(lambda info: info.getMxId() == deviceId, deviceInfos), None)
            if matchingDevice is None:
                raise RuntimeError(f"No DepthAI device found with id matching {deviceId} !")
            return matchingDevice
        elif len(deviceInfos) == 1:
            return deviceInfos[0]
        else:
            val = input("Which DepthAI Device you want to use: ")
            try:
                return deviceInfos[int(val)]
            except:
                raise ValueError("Incorrect value supplied: {}".format(val))


def showProgress(curr, max):
    """
    Print progressbar to stdout. Each call to this method will write exactly to the same line, so usually it's used as

    .. code-block:: python

        print("Staring processing")
        while processing:
            showProgress(currProgress, maxProgress)
        print(" done") # prints in the same line as progress bar and adds a new line
        print("Processing finished!")

    Args:
        curr (int): Current position on progress bar
        max (int): Maximum position on progress bar
    """
    done = int(50 * curr / max)
    sys.stdout.write("\r[{}{}] ".format('=' * done, ' ' * (50-done)) )
    sys.stdout.flush()



def downloadYTVideo(video, outputDir=None):
    """
    Downloads a video from YouTube and returns the path to video. Will choose the best resolutuion if possible.

    Args:
        video (str): URL to YouTube video
        outputDir (pathlib.Path, optional): Path to directory where youtube video should be downloaded.

    Returns:
         pathlib.Path: Path to downloaded video file

    Raises:
        RuntimeError: thrown when video download was unsuccessful
    """
    def progressFunc(stream, chunk, bytesRemaining):
        showProgress(stream.filesize - bytesRemaining, stream.filesize)

    try:
        from pytube import YouTube
    except ImportError as ex:
        raise RuntimeError("Unable to use YouTube video due to the following import error: {}".format(ex))
    path = None
    for _ in range(10):
        try:
            path = YouTube(video, on_progress_callback=progressFunc)\
                .streams\
                .order_by('resolution')\
                .desc()\
                .first()\
                .download(output_path=outputDir)
        except urllib.error.HTTPError:
            # TODO remove when this issue is resolved - https://github.com/pytube/pytube/issues/990
            # Often, downloading YT video will fail with 404 exception, but sometimes it's successful
            pass
        else:
            break
    if path is None:
        raise RuntimeError("Unable to download YouTube video. Please try again")
    return path


def cropToAspectRatio(frame, size):
    """
    Crop the frame to desired aspect ratio and then scales it down to desired size
    Args:
        frame (numpy.ndarray): Source frame that will be cropped
        size (tuple): Desired frame size (width, height)
    Returns:
         numpy.ndarray: Cropped frame
    """
    shape = frame.shape
    h = shape[0]
    w = shape[1]
    currentRatio = w / h
    newRatio = size[0] / size[1]

    # Crop width/height to match the aspect ratio needed by the NN
    if newRatio < currentRatio:  # Crop width
        # Use full height, crop width
        newW = (newRatio/currentRatio) * w
        crop = int((w - newW) / 2)
        return frame[:, crop:w-crop]
    else:  # Crop height
        # Use full width, crop height
        newH = (currentRatio/newRatio) * h
        crop = int((h - newH) / 2)
        return frame[crop:h-crop, :]


def resizeLetterbox(frame, size):
    """
    Transforms the frame to meet the desired size, preserving the aspect ratio and adding black borders (letterboxing)
    Args:
        frame (numpy.ndarray): Source frame that will be resized
        size (tuple): Desired frame size (width, height)
    Returns:
         numpy.ndarray: Resized frame
    """
    border_v = 0
    border_h = 0
    if (size[1] / size[0]) >= (frame.shape[0] / frame.shape[1]):
        border_v = int((((size[1] / size[0]) * frame.shape[1]) - frame.shape[0]) / 2)
    else:
        border_h = int((((size[0] / size[1]) * frame.shape[0]) - frame.shape[1]) / 2)
    frame = cv2.copyMakeBorder(frame, border_v, border_v, border_h, border_h, cv2.BORDER_CONSTANT, 0)
    return cv2.resize(frame, size)


def createBlankFrame(width, height, rgb_color=(0, 0, 0)):
    """
    Create new image(numpy array) filled with certain color in RGB

    Args:
        width (int): New frame width
        height (int): New frame height
        rgb_color (tuple, Optional): Specify frame fill color in RGB format (default (0,0,0) - black)

    Returns:
         numpy.ndarray: New frame filled with specified color
    """
    # Create black blank image
    image = np.zeros((height, width, 3), np.uint8)

    # Since OpenCV uses BGR, convert the color first
    color = tuple(reversed(rgb_color))
    # Fill image with color
    image[:] = color

    return image



import sys
from pathlib import Path
import blobconverter
from difflib import get_close_matches
import depthai as dai


class BlobManager:
    """
    Manager class that handles MyriadX blobs.

    """

    _zooName = None
    _zooDir = None
    _blobPath = None
    _configPath = None
    _useZoo = False
    _useBlob = False
    _zooModels = []

    def __init__(self, blobPath:Path=None, configPath:Path=None, zooName:str=None, zooDir:Path=None, progressFunc=None):
        """
        Args:
            blobPath (pathlib.Path, Optional): Path to the compiled MyriadX blob file
            configPath (pathlib.Path, Optional): Path to model config file that is used to download the model
            zooName (str, Optional): Model name to be taken from model zoo
            zooDir (pathlib.Path, Optional): Path to model zoo directory
            progressFunc (func, Optional): Custom method to show download progress, should accept two arguments - current bytes and max bytes.
        """
        if progressFunc is not None:
            blobconverter.set_defaults(progress_func=progressFunc)

        if blobPath is not None:
            self._blobPath = blobPath
            self._useBlob = True

        if zooDir is not None:
            self._zooDir = zooDir
            self._zooModels = [f.stem for f in zooDir.iterdir() if f.is_dir()]

        if zooName is not None:
            self._zooName = zooName
            if self._zooDir is not None:
                model_yml_path = self._zooDir / self._zooName / "model.yml"
                blob_paths = list((self._zooDir / self._zooName).glob("*.blob"))
                if model_yml_path.exists():
                    self._configPath = model_yml_path
                elif len(blob_paths) > 0:
                    self._blobPath = blob_paths[0]
                    self._useBlob = True
                else:
                    self._useZoo = True
            else:
                self._useZoo = True

        if configPath is not None:
            self._configPath = configPath


    def getBlob(self, shaves:int, openvinoVersion: dai.OpenVINO.Version, zooType:str = None):
        """
        This function is responsible for returning a ready to use MyriadX blob once requested.
        It will compile the model automatically using our online blobconverter tool. The compilation process will be
        ran only once, each subsequent call will return a path to previously compiled blob

        Args:
            shaves (int): Specify how many shaves the model will use. Range 1-16
            openvinoVersion (depthai.OpenVINO.Version): OpenVINO version which will be used to compile the MyriadX blob
            zooType (str, Optional): Specifies model zoo type to download blob from

        Returns:
            pathlib.Path: Path to compiled MyriadX blob

        Raises:
            SystemExit: If model name is not found in the zoo, this method will print all available ones and terminate
            RuntimeError: If conversion failed with unknown status
            Exception: If some unknown error will occur (reraise)
        """
        version = openvinoVersion.name.replace("VERSION_", "").replace("_", ".")
        if self._useBlob:
            return self._blobPath
        elif self._useZoo:
            try:
                self._blobPath = blobconverter.from_zoo(
                    name=self._zooName,
                    shaves=shaves,
                    version=version,
                    zoo_type=zooType
                )
                self._useBlob = True
                return self._blobPath
            except Exception as e:
                if hasattr(e, 'response') and hasattr(e.response, 'status_code'):
                    if "not found in model zoo" in e.response.text:
                        allModels = set(self._zooModels + blobconverter.zoo_list())
                        suggested = get_close_matches(self._zooName, allModels)
                        if len(suggested) > 0:
                            print("Model {} not found in model zoo. Did you mean: {} ?".format(self._zooName, " / ".join(suggested)), file=sys.stderr)
                        else:
                            print("Model {} not found in model zoo", file=sys.stderr)
                        raise SystemExit(1)
                    raise RuntimeError("Blob conversion failed with status {}! Error: \"{}\"".format(e.response.status_code, e.response.text))
                else:
                    raise
        elif self._configPath is not None:
            name = self._configPath.parent.stem
            self._blobPath = blobconverter.from_config(
                name=name,
                path=self._configPath,
                version=version,
                data_type="FP16",
                shaves=shaves,
            )
            self._useBlob = True
            return self._blobPath



import math
from datetime import timedelta
from queue import Queue

import cv2
import depthai as dai

from ..previews import Previews, MouseClickTracker
import numpy as np


class PreviewManager:
    """
    Manager class that handles frames and displays them correctly.
    """

    #: dict: Contains name -> frame mapping that can be used to modify specific frames directly
    frames = {}

    def __init__(self, display=[], nnSource=None, colorMap=None, depthConfig=None, dispMultiplier=255/96, mouseTracker=False, decode=False, fpsHandler=None, createWindows=True):
        """
        Args:
            display (list, Optional): List of :obj:`depthai_sdk.Previews` objects representing the streams to display
            mouseTracker (bool, Optional): If set to :code:`True`, will enable mouse tracker on the preview windows that will display selected pixel value
            fpsHandler (depthai_sdk.fps.FPSHandler, Optional): if provided, will use fps handler to modify stream FPS and display it
            nnSource (str, Optional): Specifies NN source camera
            colorMap (cv2 color map, Optional): Color map applied on the depth frames
            decode (bool, Optional): If set to :code:`True`, will decode the received frames assuming they were encoded with MJPEG encoding
            dispMultiplier (float, Optional): Multiplier used for depth <-> disparity calculations (calculated on baseline and focal)
            depthConfig (depthai.StereoDepthConfig, optional): Configuration used for depth <-> disparity calculations
            createWindows (bool, Optional): If True, will create preview windows using OpenCV (enabled by default)
        """
        self.nnSource = nnSource
        if colorMap is not None:
            self.colorMap = colorMap
        else:
            self.colorMap = cv2.applyColorMap(np.arange(256, dtype=np.uint8), cv2.COLORMAP_JET)
            self.colorMap[0] = [0, 0, 0]
        self.decode = decode
        self.dispMultiplier = dispMultiplier
        self._depthConfig = depthConfig
        self._fpsHandler = fpsHandler
        self._mouseTracker = MouseClickTracker() if mouseTracker else None
        self._display = display
        self._createWindows = createWindows
        self._rawFrames = {}

    def collectCalibData(self, device):
        """
        Collects calibration data and calculates :attr:`dispScaleFactor` accordingly

        Args:
            device (depthai.Device): Running device instance
        """

        calib = device.readCalibration()
        eeprom = calib.getEepromData()
        leftCam = calib.getStereoLeftCameraId()
        if leftCam != dai.CameraBoardSocket.AUTO:
            camInfo = eeprom.cameraData[leftCam]
            self.baseline = abs(camInfo.extrinsics.specTranslation.x * 10)  # cm -> mm
            self.fov = calib.getFov(calib.getStereoLeftCameraId())
            self.focal = (camInfo.width / 2) / (2. * math.tan(math.radians(self.fov / 2)))
        else:
            print("Warning: calibration data missing, using OAK-D defaults")
            self.baseline = 75
            self.fov = 71.86
            self.focal = 440
        self.dispScaleFactor = self.baseline * self.focal

    def createQueues(self, device, callback=None):
        """
        Create output queues for requested preview streams

        Args:
            device (depthai.Device): Running device instance
            callback (func, Optional): Function that will be executed with preview name once preview window was created
        """
        self.outputQueues = []
        for name in self._display:
            if self._createWindows:
                cv2.namedWindow(name)
            if callable(callback):
                callback(name)
            if self._createWindows and self._mouseTracker is not None:
                cv2.setMouseCallback(name, self._mouseTracker.selectPoint(name))
            if name not in (Previews.disparityColor.name, Previews.depth.name):  # generated on host
                self.outputQueues.append(device.getOutputQueue(name=name, maxSize=1, blocking=False))

        if Previews.disparityColor.name in self._display and Previews.disparity.name not in self._display:
            self.outputQueues.append(device.getOutputQueue(name=Previews.disparity.name, maxSize=1, blocking=False))
        if Previews.depth.name in self._display and Previews.depthRaw.name not in self._display:
            self.outputQueues.append(device.getOutputQueue(name=Previews.depthRaw.name, maxSize=1, blocking=False))

    def closeQueues(self):
        """
        Closes output queues for requested preview streams
        """

        for queue in self.outputQueues:
            queue.close()

    def _processFrame(self, frame, queueName):
        if self._fpsHandler is not None:
            self._fpsHandler.tick(queueName)
        return frame

    def _addRawFrame(self, frame, packet, name):
        if name in self._display:
            self._rawFrames[name] = frame

        if self._mouseTracker is not None:
            if name == Previews.disparity.name:
                rawFrame = packet.getFrame() if not self.decode else cv2.imdecode(packet.getData(), cv2.IMREAD_GRAYSCALE)
                self._mouseTracker.extractValue(Previews.disparity.name, rawFrame)
                self._mouseTracker.extractValue(Previews.disparityColor.name, rawFrame)
            if name == Previews.depthRaw.name:
                rawFrame = packet.getFrame()  # if not self.decode else cv2.imdecode(packet.getData(), cv2.IMREAD_UNCHANGED) TODO uncomment once depth encoding is possible
                self._mouseTracker.extractValue(Previews.depthRaw.name, rawFrame)
                self._mouseTracker.extractValue(Previews.depth.name, rawFrame)
            else:
                self._mouseTracker.extractValue(name, frame)

        if name == Previews.disparity.name and Previews.disparityColor.name in self._display:
            if self._fpsHandler is not None:
                self._fpsHandler.tick(Previews.disparityColor.name)
            self._rawFrames[Previews.disparityColor.name] = Previews.disparityColor.value(frame, self)

        if name == Previews.depthRaw.name and Previews.depth.name in self._display:
            if self._fpsHandler is not None:
                self._fpsHandler.tick(Previews.depth.name)
            self._rawFrames[Previews.depth.name] = Previews.depth.value(frame, self)


    def prepareFrames(self, blocking=False, callback=None):
        """
        This function consumes output queues' packets and parses them to obtain ready to use frames.
        To convert the frames from packets, this manager uses methods defined in :obj:`depthai_sdk.previews.PreviewDecoder`.

        Args:
            blocking (bool, Optional): If set to :code:`True`, will wait for a packet in each queue to be available
            callback (func, Optional): Function that will be executed once packet with frame has arrived
        """
        for queue in self.outputQueues:
            if blocking:
                packet = queue.get()
            else:
                packet = queue.tryGet()
            if packet is not None:
                frame = getattr(Previews, queue.getName()).value(packet, self)
                if frame is None:
                    print("[WARNING] Conversion of the {} frame has failed! (None value detected)".format(queue.getName()))
                    continue
                frame = self._processFrame(frame, queue.getName())
                if queue.getName() in self._display:
                    if callback is not None:
                        callback(frame, queue.getName())
                self._addRawFrame(frame, packet, queue.getName())

        for name in self._rawFrames:
            newFrame = self._rawFrames[name].copy()
            if name == Previews.depthRaw.name:
                newFrame = cv2.normalize(newFrame, None, 255, 0, cv2.NORM_INF, cv2.CV_8UC1)
            self.frames[name] = newFrame

    def showFrames(self, callback=None):
        """
        Displays stored frame onto preview windows.

        Args:
            callback (func, Optional): Function that will be executed right before :code:`cv2.imshow`
        """
        for name, frame in self.frames.items():
            if self._mouseTracker is not None:
                point = self._mouseTracker.points.get(name)
                value = self._mouseTracker.values.get(name)
                if point is not None:
                    cv2.circle(frame, point, 3, (255, 255, 255), -1)
                    cv2.putText(frame, str(value), (point[0] + 5, point[1] + 5), cv2.FONT_HERSHEY_TRIPLEX, 0.5, (0, 0, 0), 4, cv2.LINE_AA)
                    cv2.putText(frame, str(value), (point[0] + 5, point[1] + 5), cv2.FONT_HERSHEY_TRIPLEX, 0.5, (255, 255, 255), 1, cv2.LINE_AA)
            if self._fpsHandler is not None:
                self._fpsHandler.drawFps(frame, name)
            if callable(callback):
                newFrame = callback(frame, name)
                if newFrame is not None:
                    frame = newFrame
            if self._createWindows:
                cv2.imshow(name, frame)

    def has(self, name):
        """
        Determines whether manager has a frame assigned to specified preview

        Returns:
            bool: :code:`True` if contains a frame, :code:`False` otherwise
        """
        return name in self.frames

    def get(self, name):
        """
        Returns a frame assigned to specified preview

        Returns:
            numpy.ndarray: Resolved frame, will default to :code:`None` if not present
        """
        return self.frames.get(name, None)


class SyncedPreviewManager(PreviewManager):
    """
    Extension of the regular PreviewManager that allows to display all of the frames in sync
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._seqPackets = {}
        self._packetsQ = None
        self._lastSeqs = {}
        self._syncedPackets = {}
        self.nnSyncSeq = None

    def __get_next_seq_packet(self, seqKey, name, defaultPacket):
        if seqKey not in self._seqPackets:
            return defaultPacket
        elif name not in self._seqPackets:
            return self.__get_next_seq_packet(seqKey + 1, name, defaultPacket)
        else:
            return self._seqPackets[seqKey][name]

    def prepareFrames(self, blocking=False, callback=None):
        """
        This overridden function serves the same purpose - to prepare ready to use frames - but before it provide any data
        it will sync all of the packets first, using their sequence number. So any frames retrievable from this class, after
        this method is called, will be in sync with each other.

        Args:
            blocking (bool, Optional): If set to :code:`True`, will wait for a packet in each queue to be available
            callback (func, Optional): Function that will be executed once packet with frame has arrived
        """

        for queue in self.outputQueues:
            if blocking:
                packet = queue.get()
            else:
                packet = queue.tryGet()
            if packet is not None:
                seq = packet.getSequenceNum()
                packets = self._seqPackets.get(seq, {})
                packets[queue.getName()] = packet
                self._seqPackets[seq] = packets
                self._lastSeqs[queue.getName()] = seq

                if len(packets) == len(self.outputQueues):
                    self._packetsQ = Queue(maxsize=100)
                    unsyncedSeq = sorted(list(filter(lambda itemSeq: itemSeq < seq, self._seqPackets.keys())))
                    for seqKey in unsyncedSeq:
                        unsynced = {
                            synced_name: self.__get_next_seq_packet(seqKey, synced_name, synced_packet)
                            for synced_name, synced_packet in packets.items()
                        }
                        self._packetsQ.put(unsynced)
                        del self._seqPackets[seqKey]

        if self._packetsQ is not None:
            try:
                packets = self._packetsQ.get_nowait()
            except:
                packets = None
            if packets is not None:
                self.nnSyncSeq = min(map(lambda packet: packet.getSequenceNum(), packets.values()))
                for name, packet in packets.items():
                    frame = getattr(Previews, name).value(packet, self)
                    if frame is None:
                        print("[WARNING] Conversion of the {} frame has failed! (None value detected)".format(name))
                        continue
                    frame = self._processFrame(frame, name)
                    if callback is not None:
                        callback(frame, name)
                    self._addRawFrame(frame, packet, name)

        for name in self._rawFrames:
            newFrame = self._rawFrames[name].copy()
            if name == Previews.depthRaw.name:
                newFrame = cv2.normalize(newFrame, None, 255, 0, cv2.NORM_INF, cv2.CV_8UC1)
            self.frames[name] = newFrame



import io

from setuptools import setup

with open('requirements.txt') as f:
    required = f.readlines()

setup(
    name='depthai-sdk',
    version='1.1.8',
    description='This package contains convenience classes and functions that help in most common tasks while using DepthAI API',
    long_description=io.open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url='https://github.com/luxonis/depthai/tree/main/depthai_sdk',
    keywords="depthai sdk utils managers previews helpers",
    author='Luxonis',
    author_email='support@luxonis.com',
    license='MIT',
    packages=['depthai_sdk'],
    package_dir={"": "src"},  # https://stackoverflow.com/a/67238346/5494277
    install_requires=required,
    include_package_data=True,
    project_urls={
        "Bug Tracker": "https://github.com/luxonis/depthai/issues",
        "Source Code": "https://github.com/luxonis/depthai/tree/main/depthai_sdk",
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
